//
// power-utils.c
//

#include <fcntl.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/stat.h>
#include <unistd.h>

#include "power_funcs.h"
#include "pwu_config.h"
#include "un_socket.h"


extern int	pwu_notify_container(const char* msg);

#define CONTAINER_STOP	"container_stop"
#define QUIT_SERVICE	"quit"

#define CONF_FILE_PATH	"/etc/conf.d/power-utils.conf"

typedef enum PowerMode {
	PM_NONE,
	PM_SHUTDOWN,
	PM_SLEEP,
} PowerMode;


static pwu_config_t	sConfig = 0;

static int
start_containers(pwu_config_t conf)
{
	const char* target;

	// 設定ファイルで指定されたコンテナを podman_start で起動
	target = pwu_config_target_container(conf);
	if (target != NULL) {
		char	cmd[256];

		snprintf(cmd, sizeof(cmd) - 1, "/usr/bin/podman_start %s", target);
		cmd[sizeof(cmd) - 1] = '\0';
		(void)system(cmd);
	// NOTE: 少なくとも最初の版では、pod をサポートしない
	//       <- pod 内のコンテナが全て終了した時に pod の poststop hook が実行されるか
	//          否かが不明なため
	}

	return 0;
}

static void
do_sleep_and_resume()
{
	// 起床要因を設定
	(void)power_disable_wakeup_all();
	if (power_enable_wakeup_causes(sConfig) != 0) {
		printf("failed to power_enable_wakeup_causes().\n");
		return;
	}

	// OS をサスペンド
	if (power_suspend() != 0) {
		printf("failed to power_suspend().\n");
		return;
	}
	printf("back from suspend.\n");

	// サスペンドから復帰したらコンテナを起動（レジューム動作）
	start_containers(sConfig);
}

static unsigned long
get_poweroff_secs(pwu_config_t conf)
{
	const void* param;

	if (pwu_config_get_wakeup_param(conf, PWU_WAKEUP_RTC, &param) != 0
	|| (unsigned long)param < 180) {
		return 180;
	} else{
		return (unsigned long)param;
	}
}

static bool
is_live_process_of(const char* pid, const char* comm)
{
#define TASK_COMM_LEN	16  // from kernel's definition
	char	procPath[32];
	char	taskComm[TASK_COMM_LEN+1];
	int	fd;
	ssize_t	readLen;

	sprintf(procPath, "/proc/%s/comm", pid);
	fd = open(procPath, O_RDONLY);
	if (fd < 0) {
		return false;
	}
	readLen = read(fd, taskComm, sizeof(taskComm) - 1);
	(void)close(fd);
	if (readLen < 0) {
		return false;
	} else if (taskComm[readLen-1] == '\n') {
		--readLen;
	}
	taskComm[readLen] = '\0';

	return (0 == strcmp(taskComm, comm));
}

static bool
wait_swu_completion_if_progressing(void)
{
	int	fd;
	ssize_t	readLen;
	char	pid[20+1];
	struct stat	statbuf;

	fd = open("/tmp/.swupdate_lock/pid", O_RDONLY);
	if (fd < 0) {
		return false;
	}
	readLen = read(fd, pid, sizeof(pid) - 1);
	(void)close(fd);
	if (readLen < 0) {
		return false;
	} else if (pid[readLen-1] == '\n') {
		--readLen;
	}
	pid[readLen] = '\0';
	while (is_live_process_of(pid, "swupdate")) {
		if (0 == stat("/tmp/.swupdate_waiting", &statbuf)) {
			return false;  // completed but the swupdate will never exit
		}
		(void)sleep(1);
	}
	return (0 == stat("/tmp/.swupdate_rebooting", &statbuf));
}

static void
do_power_utils_service(int sockfd)
{
	unsigned char	msgBuf[128];

	for (;;) {
		size_t	recvSize = sock_recv(sockfd, msgBuf, sizeof(msgBuf));

		if (recvSize == 0 || recvSize == (size_t)-1) {
			break;
		}
		msgBuf[recvSize] = '\0';
		if (strcmp((char*)msgBuf, CONTAINER_STOP) == 0) {
			pwu_config_t	conf;

			printf("the container is stopped.\n");
			if (wait_swu_completion_if_progressing()) {
				printf("swupdate in progress, ignore stop notification.\n");
				continue;
			}
			// 設定ファイルの設定項目 'targets 'で指定したコンテナ全てが停止したかどうかを判定
				// NOTE: この版では、target を単一とする
			if (pwu_config_load(CONF_FILE_PATH, &conf) == 0) {  // 設定ファイルを再ロード
				pwu_config_dispose(sConfig);
				sConfig = conf;
			}
			switch (pwu_config_mode(conf)) {
			case PWU_MODE_NONE:
				// do nothing
				break;
			case PWU_MODE_SHUTDOWN:
				(void)power_shutdown(get_poweroff_secs(conf));
				break;
			case PWU_MODE_SLEEP:
				do_sleep_and_resume();
				break;
			case PWU_MODE_UNKNOWN:
				break;
			}
		} else if (strcmp((char*)msgBuf, QUIT_SERVICE) == 0) {
			printf("quit the service...\n");
			break;
		}
	}
}

int
main(int argc, char** argv)
{
	char*	name = strrchr(argv[0], '/');
	char*	curs;

	if (name == NULL) {
		name = argv[0];
	} else {
		++name;
	}
	curs = strchr(name, '.');
	if (curs != NULL) {
		*curs = '\0';
	}
	if (strcmp(name, "pwu_notify") == 0) {
		const char* msg;

		if (argc != 2) {
			printf("Usage: pwu_notify <message>\n");
			return -1;
		} else {
			msg = argv[1];
		}
		if ((strcmp(msg, CONTAINER_STOP) != 0)
		&& (strcmp(msg, QUIT_SERVICE) != 0)) {
			printf("unknown message: '%s'\n", msg);
			return -1;
		}
		pwu_notify_container(argv[1]);
	} else {
		int	sockfd;

		if (pwu_config_load(CONF_FILE_PATH, &sConfig) != 0) {
			printf("failed to pwu_config_load().\n");
			return -1;
		}
		sockfd = create_server_socket();
		if (sockfd < 0) {
			printf("failed to create_server_socket().\n");
			return -1;
		}
		do_power_utils_service(sockfd);
		dispose_socket(sockfd);
		pwu_config_dispose(sConfig);
	}

	return 0;
}

//
// End of File
//
